# Ben Smith: Python code for Final Project
# CHE210D

#import python modules
import numpy as np
import time

#import compiled fortran library
import finalproject_lib

import sys
import os

#whether or not to use the 3d visualization
#UseVisual = False
#if UseVisual:
    #import custom visualization library
    #import atomvis
    
import atomwrite

#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5
#set the random number seed; useful for debugging
np.random.seed(342324)

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.


def CalcEnergy(Pos):
    """Energy calculation.
Input:
    Pos: (N,3) position array
Output:
    PEnergy: potential energy, float
"""
    global M, k, L, Cut
    return finalproject_lib.calcenergy(Pos, M, k, L, Cut)
    
def CalcElectrostatics(Pos):
    """Energy calculation for electrostatic repulsion.
Input:
    Pos: (N,3) position array
Output:
    CEnergy: coulombic repulsion, float
"""
    global M, debye, L, Cut
    return finalproject_lib.calcelectrostatics(Pos, M, debye, L, Cut)


def CalcEnergyForces(Pos, Forces):
    """Energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global M, k, L, Cut
    return finalproject_lib.calcenergyforces(Pos, M, k, L, Cut, Forces)


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = Dir / np.sqrt(np.sum(Dir * Dir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [CalcEnergy(Pos + NormDir * x) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = CalcEnergy(Pos + NormDir * dx)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [CalcEnergy(Pos + NormDir * 2. * dx)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [CalcEnergy(Pos + NormDir * Dists[-1])]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = CalcEnergy(Pos + NormDir * d3)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = CalcEnergy(Pos + NormDir * d3)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    #if UseVisual:
        #if atomvis.Initialized:
            #update the positions
           # atomvis.Update(PosMin)
       # else:
            #initialize the visualization window
           #atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    PE, Forces = CalcEnergyForces(Pos, np.zeros_like(Pos))
    Dir = Forces
    OldPE = 1.e300
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = CalcEnergyForces(Pos, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N, L):
    """Returns an array of initial positions of each atom,
placed on a cubic lattice for convenience.
Input:
    N: number of atoms
    L: box length
Output:
    Pos: (N,3) array of positions
"""
    #make the position array
    Pos = np.zeros((N,3), float)
    #compute integer grid # of locations for cubic lattice
    NLat = int(N**(1./3.) + 1.)
    LatSpac = L / NLat
    #make an array of lattice sites
    r = LatSpac * np.arange(NLat, dtype=float) - 0.5*L
    #loop through x, y, z positions in lattice until done
    #for every atom in the system
    i = 0
    for x in r:
        for y in r:
            for z in r:
                Pos[i] = np.array([x,y,z], float)
                #add a random offset to help initial minimization
                Offset = 0.1 * LatSpac * (np.random.rand(3) - 0.5)
                Pos[i] = Pos[i] + Offset
                i += 1
                #if done placing atoms, return
                if i >= N:
                    return Pos
    return Pos


def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #get the acceleration from the forces
    PEnergy, Accel = CalcEnergyForces(Pos, np.zeros_like(Pos))
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))
    
def Histogram(Pos, L, M, count, Bins, Dim, N):
    return finalproject_lib.histogram(Pos, L, M, count, Bins, Dim, N)

def Radialdistribution(count, L, N, NObs):
    delta=L/float(Bins)
    gr=np.zeros(Bins,dtype=float)
    for i in range(Bins):
        gr[i]=(2.*count[i]/(float(NObs*N*(N-1))))*((3*L**3)/(4.*3.14159*(delta**3.)*((i+1)**3-i**3)))
    return gr

#### YOUR CODE HERE ####

N = 200 #number of atoms
k = 3000 #harmonic spring constant
L = (N/0.01)**(1./3.) #density
EFracTolLS = 1.e-8
EFracTolCG = 1.e-10
dx = 0.001
T = 1.0 #temperature
rc = 2.5 #cutoff
dt = 0.001
Cut = 2.5
M = [6] # computed each M separately, M = 2,4,6,8,12,16 (my loop calculated wrong values)
TEnergy = []
PdbFreq = 100
Names=range(N)
Pdb = atomwrite.pdbfile("anim.pdb", L, AtomNames = Names)
Bins = 500
HistFreq=200
count=np.zeros(Bins,dtype=int)

# assign atom names
for a in Names:
    if (a % 6 == 0): #assigned terminal atoms as negatively charged 
        Names[a]='H'
    else:
        Names[a]='C' #all other atoms are neutral
    
for x in M:
    time = 0.
    step = 0.
    NObs = 0.
    Pos = InitPositions(N, L)
    Vel = InitVelocities(N, T)
    Accel = InitAccel(Pos)
    PE, Pos = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG)
    while time <=50:
        if step % 100 == 0:   #rescale every 100 steps
            Vel = RescaleVelocities(Vel, T)    
            Pdb.write(Pos)
            step = 0.
        Pos, Vel, Accel, KEnergy, PEnergy = finalproject_lib.vvintegrate(Pos, Vel, Accel, x, k, L, rc, dt)
        if (step % HistFreq == 0):
            NObs = NObs + 1
            Histogram(Pos, L, M, count, Bins, 3, N)
        time += 0.001
        step += 1
        print "%i  %.3f %.4f  %.4f" % (x, time, KEnergy, PEnergy)
gr=Radialdistribution(count, L, N, NObs)
print "gr:"
for i in range(Bins):
    print gr[i]
Pdb.close()
    #while 2. < time <= 3.:
        #delt = [0.0001] # computed each delta T separately, T = 0.0001, 0.0002, . . . 
        #for j in delt:
           # Pos, Vel, Accel, KEnergy, PEnergy = ex3lib.vvintegrate(Pos, Vel, Accel, x, L, rc, delt)
            #TEnergy.append(KEnergy + PEnergy)
            #time += j
    #print delt, np.std(TEnergy)**2/np.mean(TEnergy)
  
        
   
    


